"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from sageopt.coniclifts.constraints.constraint import Constraint
from sageopt.coniclifts.cones import Cone
import numpy as np


class ElementwiseConstraint(Constraint):

    _CURVATURE_CHECK_ = False

    _ELEMENTWISE_CONSTRAINT_ID_ = 0

    _ELEMENTWISE_OPS_ = ('==', '<=', '>=')

    def __init__(self, lhs, rhs, operator):
        self.id = ElementwiseConstraint._ELEMENTWISE_CONSTRAINT_ID_
        ElementwiseConstraint._ELEMENTWISE_CONSTRAINT_ID_ += 1
        self.lhs = lhs
        self.rhs = rhs
        self.initial_operator = operator
        name_str = 'Elementwise[' + str(self.id) + '] : '
        self.name = name_str
        if operator == '==':
            self.expr = (self.lhs - self.rhs).as_expr().ravel()
            if ElementwiseConstraint._CURVATURE_CHECK_ and not self.expr.is_affine():
                raise RuntimeError('Equality constraints must be affine.')
            self.operator = '=='  # now we are a linear constraint "self.expr == 0"
            self.epigraph_checked = True
        else:  # elementwise inequality.
            if operator == '>=':
                self.expr = (self.rhs - self.lhs).as_expr().ravel()
            else:
                self.expr = (self.lhs - self.rhs).as_expr().ravel()
            if ElementwiseConstraint._CURVATURE_CHECK_ and not all(self.expr.is_convex()):
                raise RuntimeError('Cannot canonicalize.')
            self.operator = '<='  # now we are a convex constraint "self.expr <= 0"
            self.epigraph_checked = False

    def variables(self):
        return self.expr.variables()

    def is_affine(self):
        if self.operator in ['==']:
            return True
        else:
            return self.expr.is_affine()

    def is_elementwise(self):
        return True

    def is_setmem(self):
        return False

    def conic_form(self):
        from sageopt.coniclifts.base import ScalarVariable
        # This function assumes that every Constraint "c" in "constraints" has
        # c.is_affine() == True. (For constraints that were not affine when first
        # constructed, we must have since performed an epigraph substitution to
        # handle the nonlinear terms.)
        #
        # The vector "K" returned by this function may only include entries for
        # the zero cone and R_+.
        #
        # Note: signs on coefficients are inverted in this function. This happens
        # because flipping signs on A and b won't affect the zero cone, and
        # it correctly converts affine constraints of the form "expression <= 0"
        # to the form "-expression >= 0". We want this latter form because our
        # primitive cones are the zero cone and R_+.
        if not self.epigraph_checked:
            raise RuntimeError('Cannot canonicalize without check for epigraph substitution.')
        m = self.expr.size
        b = np.empty(shape=(m,))
        if self.operator == '==':
            K = [Cone('0', m)]
        elif self.operator == '<=':
            K = [Cone('+', m)]
        else:
            raise RuntimeError('Unknown operator.')
        A_rows, A_cols, A_vals = [], [], []
        for i, se in enumerate(self.expr.flat):
            if len(se.atoms_to_coeffs) == 0:
                b[i] = -se.offset
                A_rows.append(i)
                A_cols.append(int(ScalarVariable.curr_variable_count())-1)
                A_vals.append(0)  # make sure scipy infers correct dimensions later on.
            else:
                b[i] = -se.offset
                A_rows += [i] * len(se.atoms_to_coeffs)
                col_idx_to_coeff = [(a.id, c) for a, c in se.atoms_to_coeffs.items()]
                A_cols += [atom_id for (atom_id, _) in col_idx_to_coeff]
                A_vals += [-c for (_, c) in col_idx_to_coeff]
        return A_vals, np.array(A_rows), A_cols, b, K

    def violation(self, norm_ord=None):
        expr = (self.lhs - self.rhs).as_expr()
        expr_val = expr.value
        if self.initial_operator == '<=':
            ignore = expr_val <= 0
            expr_val[ignore] = 0
        elif self.initial_operator == '>=':
            ignore = expr_val >= 0
            expr_val[ignore] = 0
        else:
            expr_val = np.abs(expr_val)
        residual = expr_val.ravel()
        viol = np.linalg.norm(residual, ord=norm_ord)
        return viol



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
from sageopt.coniclifts.base import Expression, Variable
from sageopt.coniclifts.cones import Cone


def sum_relent(x, y, z, aux_vars):
    # represent "sum{x[i] * ln( x[i] / y[i] )} + z <= 0" in conic form.
    # return the Variable object created for all epigraphs needed in
    # this process, as well as A_data, b, and K.
    if not isinstance(x, Expression):
        x = Expression(x)
    if not isinstance(y, Expression):
        y = Expression(y)
    if not isinstance(z, Expression):
        z = Expression(z)
    x = x.ravel()
    y = y.ravel()
    z = z.ravel()
    if z.size != 1 or not z.is_affine():
        raise RuntimeError('Illegal argument to sum_relent.')
    if x.size != y.size:
        raise RuntimeError('Illegal arguments to sum_relent.')
    num_rows = 1 + 3 * x.size
    aux_var_ids = aux_vars.scalar_variable_ids
    b = np.zeros(num_rows,)
    K = [Cone('+', 1)] + [Cone('e', 3) for _ in range(x.size)]
    A_rows, A_cols, A_vals = [], [], []
    # populate the first row
    z_id2co = [(a.id, co) for a, co in z[0].atoms_to_coeffs.items()]
    A_cols += [aid for aid, _ in z_id2co]
    A_vals += [-co for _, co in z_id2co]
    A_cols += [aid for aid in aux_var_ids]
    A_vals += [-1] * len(aux_var_ids)
    A_rows += [0] * len(A_vals)
    b[0] = -z[0].offset
    # populate the epigraph terms
    for i in range(x.size):
        curr_cone_start = 1 + 3 * i
        # first entry of exp cone
        A_rows.append(curr_cone_start)
        A_cols.append(aux_var_ids[i])
        A_vals.append(-1)
        # third entry of exp cone
        id2co = [(a.id, co) for a, co in x[i].atoms_to_coeffs.items()]
        A_rows += [curr_cone_start + 2] * len(id2co)
        A_cols += [aid for aid, _ in id2co]
        A_vals += [co for _, co in id2co]
        b[curr_cone_start + 2] = x[i].offset
        # second entry of exp cone
        id2co = [(a.id, co) for a, co in y[i].atoms_to_coeffs.items()]
        A_rows += [curr_cone_start + 1] * len(id2co)
        A_cols += [aid for aid, _ in id2co]
        A_vals += [co for _, co in id2co]
        b[curr_cone_start + 1] = y[i].offset
    return A_vals, np.array(A_rows), A_cols, b, K, aux_vars


def elementwise_relent(x, y, z):
    """
    Return variables "z" and conic constraint data for the system
        x[i] * ln( x[i] / y[i] ) <= z[i]

    A_vals - a list of floats,
    np.array(A_rows) - a numpy array of ints,
    A_cols - a list of ints,
    b - a numpy 1darray,
    K - a list of coniclifts Cone objects (of type 'e'),
    z - a coniclifts Variable
    """
    if not isinstance(x, Expression):
        x = Expression(x)
    if not isinstance(y, Expression):
        y = Expression(y)
    x = x.ravel()
    y = y.ravel()
    if x.size != y.size:
        raise RuntimeError('Illegal arguments to sum_relent.')
    num_rows = 3 * x.size
    aux_var_ids = z.scalar_variable_ids
    b = np.zeros(num_rows, )
    K = [Cone('e', 3) for _ in range(x.size)]
    A_rows, A_cols, A_vals = [], [], []
    # populate the epigraph terms
    # for ECOS, the negative of the epigraph maps to the first term,
    #           the "x" maps to the third term,
    #           and the "y" maps to the second term.
    for i in range(x.size):
        curr_cone_start = 3 * i
        # first entry of exp cone
        A_rows.append(curr_cone_start)
        A_cols.append(aux_var_ids[i])
        A_vals.append(-1)
        # third entry of exp cone
        id2co = [(a.id, co) for a, co in x[i].atoms_to_coeffs.items()]
        A_rows += [curr_cone_start + 2] * len(id2co)
        A_cols += [aid for aid, _ in id2co]
        A_vals += [co for _, co in id2co]
        b[curr_cone_start + 2] = x[i].offset
        # third entry of exp cone
        id2co = [(a.id, co) for a, co in y[i].atoms_to_coeffs.items()]
        A_rows += [curr_cone_start + 1] * len(id2co)
        A_cols += [aid for aid, _ in id2co]
        A_vals += [co for _, co in id2co]
        b[curr_cone_start + 1] = y[i].offset
    return A_vals, np.array(A_rows), A_cols, b, K, z



import setuptools

setuptools.setup(
    name='sageopt',
    version='0.3.4',
    author='Riley John Murray',
    url='https://github.com/rileyjmurray/sageopt',
    author_email='rmurray@caltech.edu',
    description='Signomial and polynomial optimization via SAGE relaxations',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics'
    ],
    python_requires='>=3.5',
    install_requires=["ecos >= 2",
                      "numpy >= 1.14",
                      "scipy >= 1.1"]
)



"""
   Copyright 2019 Riley John Murray

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""


class Solver(object):

    @staticmethod
    def apply(c, A, b, K, destructive, compilation_options):
        raise NotImplementedError()

    @staticmethod
    def solve_via_data(data, params):
        raise NotImplementedError()

    @staticmethod
    def parse_result(solver_output, inv_data, var_mapping):
        raise NotImplementedError()

    @staticmethod
    def is_installed():
        raise NotImplementedError()
